// This may look like C code, but it is really -*- C++ -*-
/*
 ************************************************************************
 *
 *			   UNT Virtual Machine
 *			   Input/Output Channel
 *
 * The present file emulates operations of the I/O channel and the
 * divice driver.
 *
 * Note that all the channels are connected to the I/O bus and gets its
 * number from it (so the channel numbers tell the order the channels are
 * connected to the bus). The channel itself consists of the device-
 * independent part (that performs fetching and decoding of the CCW,
 * moves data between io_buffer and memory, handles error conditions
 * and raises the interrupt) and the device-dependent part. The latter
 * is responsible for specific for each device operations necessary to
 * transmit data between the io_buffer and the "media".
 *
 * Once started, the channel executes the given sequence of CCW (channel
 * program) until the program is over (no command chain flag set at the
 * last CCW) or some error occured. When the channel stops, 'intr_signal'
 * is raised to inform the OS. The status of the completed program is
 * available through q_status().
 *
 * Note each read/write channel operation transmit exactly one page
 * of data between the "media" and the memory. The current implementation
 * does not support address translation during I/O (and neither do most of
 * the real computing system). So, the page number specified in the
 * CCW is supposed to be the number of a page in the physical memory
 * (page frame). It implies that the OS is responsible for address
 * translation itself if necessary when constructing the channel program.
 * Channel program itself should also be in the physical memory.
 *
 ************************************************************************
 */

#pragma implementation
#include "hardware.h"
#include "myenv.h"
#include <std.h>
#include <sys/file.h>


/*
 *------------------------------------------------------------------------
 *			  I/O Bus operations
 */

IOBus::IOBus(void)
{
  no_channels = 0;
}

				// Connect a channel to the i/o bus
				// Return the channel number on the bus,
				// an integer starting from 1
int IOBus::connect_channel(IOChannel * channel)
{
  assure(++no_channels < max_no_channels, "Too many channels in the program");
  channels[no_channels - 1] = channel;
  return no_channels;
}

				// Dump the status of all the channels
void IOBus::dump(void)
{
  register int i;
  for(i=0; i<no_channels; i++)
    channels[i]->dump();
}

				// Start a channel program on a specified
				// channel. Return FALSE if the operation
				// was rejected upfront
int IOBus::start(const unsigned short channel_no, const Address ccw_ptr)
{
  if( channel_no == 0 || channel_no > no_channels )
  {
    console("Channel no. %d is not in 1..%d, can't start the channel program",
	    channel_no,no_channels);
    return 0;
  }
  channels[channel_no-1]->start(ccw_ptr);
  return 1;
}

/*
 *------------------------------------------------------------------------
 *		        Channel general initialization
 */

IOChannel::IOChannel(Memory& _memory,IOBus& io_bus)
	: memory(_memory),
	  stopped("Channel operation",0),
	  intr_signal("Channel interrupt",0)
{

  channel_no = io_bus.connect_channel(this);

  cpc = 0;
  file_handle = -1;
  rec_count = 0;
  status = OK;

  if( !newproc("Channel process",1) )
  {					// This is a channel process
    for(;;)
    {
      stopped--;			// Wait until can run again
      if( (status = execute_CCW()) == OK && current_CCW.keep_going )
	stopped++;			// If everything was fine, keep going
      else
	intr_signal++;
    }
  }
}

				// Starting a channel
void IOChannel::start(Address channel_program_ptr)
{
  if( !is_running() )
    single_message("Channel: starting the program at %6o",channel_program_ptr),
    cpc = channel_program_ptr,
    stopped++;
  else
    _error("FATAL: tried to start a running channel");
}

/*
 *------------------------------------------------------------------------
 *		   Emulates a single channel operation
 *
 * In other words, fetch a CCW at current cpc, analyse it, have the
 * device driver do its part and move the data to/from the computer main
 * memory. The program returns the STATCODE.
 * Note, that cpc always points to the CCW following the one which has been
 * executed (or failed).
 *
 */

IOChannel::STATCODE IOChannel::execute_CCW(void)
{
  if( file_handle == EOF )
  {
    console("Channel %d: file has not been open",channel_no);
    return DEVICE;
  }

  if( memory[cpc].out_of_range )
    return CCWADDR;

					// Fetch the CCW
  *(Word *)&current_CCW = memory[cpc++].contents;
  opcode = (CCWOpCodes)current_CCW.opcode;
  buf_page = current_CCW.page_no;

				// Print the trace info
  begin_printing();
  message("\nChannel %d: Executing instruction", channel_no);
  message("\n  CPC  Opcode  Chain   Page no"
	  "\n%4ob  %3d       %s       %2d",
	  cpc-1,opcode, (current_CCW.keep_going ? "Y" : "N"), buf_page);
  message("\n");
  end_printing();
				// Interpret the instruction
  switch( opcode )
  {
    case ILLEGAL_CCW:
         return ILLEGOP;

    case SETMAP:
	 if( buf_page != 0 )
	   _error("Sorry, address translation during I/O isn't implemented");
         return OK;

    case SETPTBR:
	 _error("FATAL: address translation during I/O isn't implemented");

    case SETPTL:
	 _error("FATAL: not implemented yet");

    case SEEKOP:
         return device_driver(DEV_SEEK);

    case READOP:
         {					// Read a record
	   STATCODE read_status = device_driver(DEV_READ);
	   if( read_status != OK )
	     return read_status;
	 }
	 if( put_record(buf_page) )		// Put the read record into
	   return OK;				// the memory
	 else
	   return MEMORY;

    case WRITEOP:
	 if( !get_record(buf_page) )		// Get a record to write
	   return MEMORY;
         return device_driver(DEV_WRITE);

    default:
         return ILLEGOP;
  }

  return OK;
}


/*
 *------------------------------------------------------------------------
 *		    Non-device specific channel operations
 */

				// Tell what's going on
void IOChannel::dump(void) const
{
}

				// Get a record to write from the main memory
				// to the io_buffer.
				// Return FALSE if we've got a memory problem
int IOChannel::get_record(const int page_no)
{
  register int i;
  for(i=0; i<rec_size; i++)
  {
    Memory::MemoryAnswer ans = memory[page_no*Memory::pagesize+i];
    if( ans.out_of_range )
      return 0;
    else
      io_buffer[i] = ans.contents;
  }
  return 1;
}

				// Put a read record to the main memory
				// from the io_buffer.
				// Return FALSE if we've got a memory problem
int IOChannel::put_record(const int page_no)
{
  register int i;
  for(i=0; i<rec_size; i++)
  {
    Memory::MemoryAnswer ans = memory[page_no*Memory::pagesize+i];
    if( ans.out_of_range )
      return 0;
    else
      ans.contents = io_buffer[i];
  }
  return 1;
}

				// Open the UNIX file the channel is associated
				// with
void IOChannel::associate_with_unix_file(const char * file_name)
{
  message("\nOpening the file %s\n",file_name);
  file_handle = open(file_name,O_RDWR,0);
  if( file_handle < 0 )
    perror("File opening error"),
    console("Channel %d will be unoperational due to the reason above",
	    channel_no);
  else
    console("Channel %d is operational and associated with the file '%s'",
	    channel_no,file_name);
}

/*
 *------------------------------------------------------------------------
 *			Device-specific processing
 */

CardReader::CardReader(Memory & main_memory, IOBus& io_bus)
	: IOChannel(main_memory,io_bus)
{
  in_file = (void *)0;
}

				// Card reader device driver
				// Note, the driver reads a record from
				// the file to the io_buffer until '\n' or
				// the entire record is read (whichever
				// comes first). In the former case,
				// the record is padded with blanks
IOChannel::STATCODE CardReader::device_driver(const DEVICE_OP dev_op)
{
  if( dev_op == DEV_SEEK )
    return OK;
  if( dev_op != DEV_READ )
    return ILLEGOP;

  if( in_file == (void *)0 )
    if( (in_file = fdopen(file_handle,"r")) == (void *)0 )
    {
      perror("Error opening the card reader file");
      console("Card reader (channel %d) is no longer operational",
	      q_channel_no());
      file_handle = -1;
      return DEVICE;
    }

  if( fgets((char *)io_buffer,sizeof(io_buffer),(FILE *)in_file) == (void *)0 )
    if( feof((FILE *)in_file) )
    {
      console("Card reader has nothing to read");
      return END_FILE;
    }
    else
    {
      perror("I/O error on the card reader");
      return DEVICE;
    }

  register char *p;
  if( (p = strchr((char *)io_buffer,'\n')) != 0 )
    while( p < (char *)io_buffer + sizeof(io_buffer) )
      *p++ = ' ';
  else
  {					// Read the last char into buffer
    p = (char *)io_buffer + sizeof(io_buffer) - 1;
    int c = fgetc((FILE *)in_file);
    if( c == EOF || c == '\n' )
      c = ' ';
    *p = c;
  }
  rec_count++;
  return OK;
}

				// Line printer device driver
				// Note, the driver automatically writes
				// '\n' after any write operation.
				// DEV_SEEK is interpreted as writing
				// '\n' (advance the paper)
IOChannel::STATCODE LinePrinter::device_driver(const DEVICE_OP dev_op)
{
  char new_line = '\n';

  if( dev_op == DEV_SEEK )
  {					// Seek is just advance the paper
    if( write(file_handle,&new_line,1) != 1 )
    {
      perror("Line printer file error");
      return DEVICE;
    }
    rec_count++;
    return OK;
  }

  if( dev_op != DEV_WRITE )
    return ILLEGOP;

  if( write(file_handle,(char *)io_buffer,sizeof(io_buffer)) 
      != sizeof(io_buffer) )
  {
    perror("Line printer file error");
    return DEVICE;
  }
  return device_driver(DEV_SEEK);
}


				// Hard disk device driver
IOChannel::STATCODE HardDisk::device_driver(const DEVICE_OP dev_op)
{
  assert( sizeof(io_buffer) == disk_sector_size );
  switch(dev_op)
  {
    case DEV_SEEK:
         rec_count = q_buf_page();		// It should contain the 
	 if( rec_count > hi_disk_addr )		// seek addr
	 {
	   console("HardDisk: disk sector address %d is too big",
		   rec_count);
	   return SEEKADDR;
	 }
	 if( lseek(file_handle,rec_count*disk_sector_size,0) < 0 )
	 {
	   perror("HardDisk: seek failed");
	   return DEVICE;
	 }
	 return OK;

    case DEV_READ:
	 if( read(file_handle,(char *)io_buffer,sizeof(io_buffer)) !=
	     sizeof(io_buffer) )
	 {
	   perror("HardDisk: read failed");
	   return DEVICE;
	 }
	 rec_count++;
	 return OK;

    case DEV_WRITE:
	 if( write(file_handle,(char *)io_buffer,sizeof(io_buffer)) !=
	     sizeof(io_buffer) )
	 {
	   perror("HardDisk: write failed");
	   return DEVICE;
	 }
	 rec_count++;
	 return OK;

    default:
         return ILLEGOP;
  }
  return OK;
}
