// This may look like C code, but it is really -*- C++ -*-
/*
 ************************************************************************
 *
 *			   UNT Virtual Machine
 *
 *				CPU Manager
 * Manages the most important and the leading resource in the computer.
 * It handles all the interrupts (which are CPU interrupts), and 
 * all the system gets activity through it. Implements the concept of
 * the process.
 *
 ************************************************************************
 */

#ifndef _cpu_manager_h
#define _cpu_manager_h 1
#pragma interface

#include "hardware.h"
#include "semaphores.h"
#include "mem_manager.h"

enum SVCCODE { DUMP_STATUS=1, FORK, WAIT_KIDS, SEMINIT, SEM_P, SEM_V, KILL };
enum HANDLER_STATUS { RESUME, DISPATCH };

class IOManager;

class CPUManager : public ProcessTable
{
  CentralProcessingUnit& CPU;		// CPU to manage
  MemoryManager& memory_manager;	// Manager of the memory
  IOManager& io_manager;		// Manager of I/O

  const int Time_quant = 10;		// In clock units
  Semaphore Halt;			// V(Halt) to halt the system

  PCB * running_pcb;			// Ptr to the running PCB (which is 
					// in none of the queues)

  BasicQueue readyPCBs;			// Queue of ready processes
  SemaphoreTable * semas;

				// Handlers
  HANDLER_STATUS trap_handler(const CentralProcessingUnit::TRAPCODE trap_code);
  HANDLER_STATUS svc_handler(const SVCCODE svc_code);
  HANDLER_STATUS mfault_handler(const MemoryManagementUnit::MemoryFault code);
  HANDLER_STATUS initiate_IO(const OpCodes opcode,Address IO_params);

  HANDLER_STATUS terminate();		// Terminate the current process
  void dispatch();			// Pick up a new process to run
  void prepare_for_running();		// Prepare running_pcb for running
  void save_context();			// Save the CPU context in
					// the running_pcb
  void kill(PCB & pcb);			// Kill the process
  void dump(void);			// Dump the current OS status

  					// Create a new process at EA.
  HANDLER_STATUS fork(const int rega,Address EA);	// put PID in rega
  HANDLER_STATUS wait_for_kids(void);	// Wait for kids to terminate
  HANDLER_STATUS shoot(const PID pid);	// Try to kill the process

					// Create a new semaphore with
					// initial value EA. Put SID in rega
  HANDLER_STATUS create_semaphore(const int rega,Address EA);

					// Check to see that the Sema is valid
					// to perform P or V operation on.
  Sema * valid_semaphore(const SID sid);
					// Semaphore P-operation
  HANDLER_STATUS semaphore_p(const SID sid);
					// Semaphore P-operation
  HANDLER_STATUS semaphore_v(const SID sid);

public:
  CPUManager(CentralProcessingUnit& _CPU,MemoryManager& _mem_manager,
	     IOManager& _io_manager);
  ~CPUManager(void) {}
  void commence();			// Load a program from the drum
					// and run it
  void wait_for_halt(void)		{ Halt--; }

					// Post the completion of the I/O
					// request
  void io_request_completed(const PID pid);

					// Read/write a word from/to the 
					// (virtual) address space of a
					// current process
					// Return FALSE if failed
  int read_word(const Address va, Word& dest);
  int write_word(const Address va, Word src);

					// Read/write a word from/to the 
					// PHYSICAL address space
					// Error is fatal
  void read_word_phys_memory(const Address ra, Word& dest);
  void write_word_phys_memory(const Address ra, Word src);

				// Lock the page of the running process
				// at the specified VA
  int lock_page(Address va);
				// Unlock the page of a given process
				// at the specified VA
  void unlock_page(const PID pid,Address va);


				// Translate the virtual address to physical
				// one using the translation tables of the
				// running process. The page has to be loaded
				// Returns 0 if fails
  Address translate_va(Address va); 

				// Reserve the physical page for the system
				// Returns the physical address it starts
  Address reserve_physical_page()
  	{ return memory_manager.reserve_physical_page(); }
};

#endif
