// This may look like C code, but it is really -*- C++ -*-
/*
 ************************************************************************
 *
 *			   UNT Virtual Machine
 *
 *		     High Level Virtual Memory Manager
 *
 * This is a high-level extension of the memory unit, and uses paging to
 * manage memory requests of processes.
 * Page selection is on demand, that's the page isn't loaded until it is 
 * missed by the CPU to complete the execution of an instruction.
 *
 * The present version supports the working set concept and an approximate
 * LRU page replacement strategy. Working set of a process is a set of
 * all pages that are currently referenced (mapped, i.e. loaded into
 * the physical memory) by the process. In other words, it is the set of
 * pages the process is working with. Working set quota limits the number
 * of the pages the process is allowed to have mapped. If the working set
 * is filled up to the quota, yet the process needs (demands) one more page,
 * one page of the working set should be unreferenced (and swapped out if
 * it has been modified) before the request to map new page is served. 
 * To decide which page to unmap, an approximate LRU page replacement
 * strategy is used. 
 * It means, the algorithm will try to find the least recently used page
 * of the working set and push it out. To figure out which page has been
 * used least recently, the memory context of the process keeps a time-since-
 * last-access count of the page usage. When the process gets control
 * of the CPU and its page table gets loaded into the hardware, was_ref
 * bit of every page is dropped. When the CPU accesses memory, the memory
 * hardware sets the was_ref bit of the accessed page. When the process
 * loses control of the CPU (because it runs into the blocking condition,
 * or because it exhausted its time slice), the time-since-last-access
 * count for all the pages that haven't been accessed during that
 * process run is incremented. Therefore, the larger the count, the
 * least recently the page has been accessed (used). Note, since 
 * process switchings out occure non-uniformly (the process loses
 * the CPU control not only because of the time click, but also because
 * of blocking), the LRU count is only approximately correct. Yet,
 * interrupts in the real system occur quite regularly, so this
 * approximation seems fairly accurate.
 *
 ************************************************************************
 */

#pragma implementation
#include "mem_manager.h"

#include "myenv.h"
#include <std.h>

MemoryManager::MemoryManager(Memory& _physical_memory)
	: physical_memory(_physical_memory),
	  virtual_space(_physical_memory,(Memory::hiaddr+1)/Memory::pagesize,
			Memory::no_drum_sectors)
{
  no_programs = 0;
}


				// Read the drum header and tell the
				// number of programs on the drum
int MemoryManager::q_no_programs(void)
{
  virtual_space.declare_page_private(0);
  physical_memory.drum_read(0,PageFrameTable::sys_page_frame);
  const Address base_address = PageFrameTable::sys_page_frame*Memory::pagesize;
  register Address curr_address = base_address;
  no_programs = physical_memory[curr_address++].contents;
  assert( no_programs > 0 && no_programs <= max_no_programs );
  register int i;
  for(i=0; i<no_programs; i++)
    prog_descriptors[i].pc = physical_memory[curr_address++].contents,
    prog_descriptors[i].mem_map = physical_memory[curr_address++].contents;

  return no_programs;
}


void MemoryManager::dump_status(void) const
{
  virtual_space.dump_status();
}

/*
 *------------------------------------------------------------------------
 *		       Servicing memory requests of a process
 */

				// Initializing all areas
MMContext::MMContext(void)
{
  register int i;
  for(i=0; i<process_virtual_space; i++)
    vm_map[i] = NIL_vpn;
  memset(page_table,0,sizeof(page_table));
  memset(time_since_last_ref,0,sizeof(time_since_last_ref));
  no_mapped_pages = 0;
  no_page_faults = 0;
}

				// Given vm_map, map of the process virtual
				// space, construct the page table
void MMContext::build_page_table(const VirtualPageTable& virtual_space)
{
  register int i;
  for(i=0; i<process_virtual_space; i++)
  {
    PageTableEntry& pte = page_table[i];
    pte.valid = pte.was_ref = pte.was_written = 0;
    if( vm_map[i] == NIL_vpn )		// The page hasn't been allocated
      pte.any_access = pte.read_only = pte.exec_only = 0;
    else
    {
      pte.any_access = 1;		// The page has been allocated
      const VirtualPage& vp = virtual_space[vm_map[i]];
      pte.exec_only = vp.q_data_page() ? 0 : 1;
      if( vp.page_frame != NIL_pfn)
	pte.phys_page_no = vp.page_frame,// The page is mapped
	pte.valid = 1;
    }
  }
}


				// Load the context to the MMU
				// Note, the referenced bit for every page
				// in the page table is cleared to
				// find out which pages are going to be
				// accessed during the time the process
				// is running on CPU
void MMContext::load_MMU(MemoryManagementUnit& mmu)
{
  memcpy((MMUContext *)&mmu,this,sizeof(MMUContext));
					// Load the page table
  const Address base_address = PageFrameTable::sys_page_frame*Memory::pagesize;
  assert( page_table_len <= process_virtual_space );
  assert( page_table_len <= Memory::pagesize );

  register int i;
  for(i=0; i<page_table_len; i++)
  {
    PageTableEntry& pte = page_table[i];
    pte.was_ref = 0;			// Clear the ref bit
    *(PageTableEntry *)&(mmu.memory[base_address+i].contents) = pte;
  }
  mmu.clear_error();
}

				// Save the MMU context
				// Was_written/was_referenced fields in the
				// page_table might have been changed. This
				// needs to be saved. The referenced bit 
				// is used to find out which pages have been
				// actually accessed, and to increment counter
				// for the pages that haven't been used
void MMContext::save_MMU(const MemoryManagementUnit& mmu)
{
					// Save the page table
  const Address base_address = PageFrameTable::sys_page_frame*Memory::pagesize;
  assert( page_table_len <= process_virtual_space );
  assert( page_table_len <= Memory::pagesize );

  register int i;
  for(i=0; i<page_table_len; i++)
  {
    const PageTableEntry& pte =
      *(PageTableEntry *)&(mmu.memory[base_address+i].contents);
    if( pte.valid && !pte.was_ref )
      time_since_last_ref[i]++;
    page_table[i] = pte;
  }
}

				// Dump whatever's in there
void MMContext::dump(void) const
{
  if( !enabled_address_translation )
  {
    message("\nVirtual memory is disabled for this process\n");
    return;
  }

  message("\nProcess virtual space is %d pages long",page_table_len);
  message("\nWorking set size %d, working set quota %d\n",no_mapped_pages,
	  working_set_quota);
  message("\nAddress range\tAccess\tMapped to   Drum sector"
	  "\tWritten\tReferenced LRU time\n");
  register int i;
  for(i=0; i<page_table_len; i++)
  {
    const PageTableEntry& pte = page_table[i];
    if( !(pte.any_access || pte.read_only || pte.exec_only) )
      continue;					// Page isn't allocated
    message("%06o-%06o \t %s\t",Memory::pagesize*i,Memory::pagesize*(i+1)-1,
	    pte.exec_only ? "Exec" : pte.read_only ? "Read" : "Any");
    if( pte.valid )
      message("%06o  ",pte.phys_page_no*Memory::pagesize);
    else
      message("Not Mapped ");
    message("\t%d\t   %s\t    %s",vm_map[i],pte.was_written ? "Y" : "N",
	    pte.was_ref ? "Y" : "N");
    if( pte.valid )
      message("         %d\n",time_since_last_ref[i]);
    else
      message("\n");
  }
  message("\nProcess has %d page faults\n\n",no_page_faults);
}

					// Load program and return starting PC
Address MemoryManager::load_program
	(MMContext& context,const unsigned int prog_no)
{
  assure(prog_no < (unsigned)no_programs, 
	 "FATAL: an attempt to load nonexistent program");
  ProgDescr& prog_descr = prog_descriptors[prog_no];
					// Read the memory map from the drum
  virtual_space.declare_page_private((VPN)prog_descr.mem_map);
  physical_memory.drum_read((VPN)prog_descr.mem_map,
			    PageFrameTable::sys_page_frame);
					// Set up the vm_map and page_table
  const Address base_address = PageFrameTable::sys_page_frame*Memory::pagesize;
  register int i;
  for(i=0; i<MMContext::process_virtual_space; i++)
  {
    assert( i < Memory::pagesize );
    int drum_sector_no = physical_memory[base_address+i].contents;
    if( drum_sector_no == -1 )
      context.vm_map[i] = NIL_vpn;
    else
    {
      virtual_space.declare_page_private((VPN)drum_sector_no);
      if( !is_text_segment(i) )
	virtual_space[(VPN)drum_sector_no].declare_data_page();
      context.vm_map[i] = drum_sector_no;
    }
  }

  context.build_page_table(virtual_space);
  context.no_page_faults = 0;
  context.no_mapped_pages = 0;
  memset(context.time_since_last_ref,0,sizeof(context.time_since_last_ref));

  context.page_table_addr = PageFrameTable::sys_page_frame*Memory::pagesize;
  context.page_table_len  = MMContext::process_virtual_space;
  context.enabled_address_translation = 1;

  return prog_descr.pc;
}

				// Take care of son's virtual space
				// Return FALSE if we've got some problem
int MemoryManager::fork_son
	(MMContext& son_context,const MMContext& dad_context)
{
					// Check is everything's in real memory
  if( !dad_context.enabled_address_translation )
  {
    son_context.enabled_address_translation = 0;
    return 1;
  }
  					// Set up the vm_map
  register int i;
  son_context.no_mapped_pages = 0;
  for(i=0; i<MMContext::process_virtual_space; i++)
  {
    VPN dad_vpn = dad_context.vm_map[i];
    if( dad_vpn == NIL_vpn )
    {
      son_context.vm_map[i] = NIL_vpn;
      continue;
    }

    const VirtualPage& dad_vp = virtual_space[dad_vpn];
    if( !dad_vp.q_data_page() )
    {					// Code text page, to be shared
      virtual_space.allocate(dad_vpn);
      son_context.vm_map[i] = dad_vpn;
      if( dad_vp.is_mapped() )		// Son would also refer to the
	virtual_space.reference(dad_vpn),	// dad's mapped pages
        son_context.no_mapped_pages++;
      continue;
    }

    VPN son_vpn = virtual_space.allocate();
    if( son_vpn == NIL_vpn )
    {
      console("No virtual space for the son process. It has to be killed");
      return 0;
    }

    son_context.vm_map[i] = son_vpn;
    virtual_space[son_vpn].declare_data_page();
    virtual_space.copy(son_vpn,dad_vpn);
  }

  son_context.build_page_table(virtual_space);
  son_context.no_page_faults = 0;
  memset(son_context.time_since_last_ref,0,
	 sizeof(son_context.time_since_last_ref));

  son_context.page_table_addr = 
    PageFrameTable::sys_page_frame*Memory::pagesize;
  son_context.page_table_len  = MMContext::process_virtual_space;
  son_context.enabled_address_translation = 1;

  return 1;
}

				// Serve the request to bring a virtual page
				// into the real memory
MemoryManager::MMAnswer MemoryManager::load_the_page
	(MMContext& context,MMUContext::VirtualMemoryAddr culprit_VA)
{
  unsigned int page_no = culprit_VA.page_no;
  assert( page_no < (unsigned)context.page_table_len );
  VPN vpn = context.vm_map[page_no];

  if( context.no_mapped_pages >= context.working_set_quota )
    replace_LRU_page(context);

  PFN pfn = virtual_space.reference(vpn);
  if( pfn == NIL_pfn )
    return PhysMemoryExhausted;
  context.no_mapped_pages++;
  VirtualPage& vp = virtual_space[context.vm_map[page_no]];
  
  if( !vp.is_shared_mapped() )
    context.no_page_faults++;		// The page has been physically loaded
  else					// The page might have been loaded by
    if( vp.q_data_page() )		// siblings (if it's shared)
      _error("Non-shared data page %d turns out to be loaded into frame %d",
	     vp.id,vp.page_frame);

  MMUContext::PageTableEntry& pte = context.page_table[page_no];
  pte.phys_page_no = vp.page_frame;
  pte.valid = 1;
  context.time_since_last_ref[page_no] = 0;
  return Ok;
}

					// Release the virtual memory
					// occupied by the process
void MemoryManager::dispose(MMContext& context)
{
  if( !context.enabled_address_translation )
    return;				// No virtual memory was occupied
  register int i;
  for(i=0; i<MMContext::process_virtual_space; i++)
  {
    VPN vpn = context.vm_map[i];
    if( vpn == NIL_vpn )
      continue;
    MMUContext::PageTableEntry& pte = context.page_table[i];
    if( pte.valid )
      virtual_space.unreference(vpn,0),
      context.no_mapped_pages--;
    virtual_space.dispose(vpn);
    context.vm_map[i] = NIL_vpn;
  }
  assert(context.no_mapped_pages == 0);
  memset(context.time_since_last_ref,0,sizeof(context.time_since_last_ref));
  context.no_page_faults = 0;
  context.page_table_addr = 0;
  context.page_table_len  = 0;
  context.enabled_address_translation = 0;
}

				// Swap out a victim - forcibly deprive it
				// of mapped pages
				// Return 0 if no page frame was released
				// Swapping goals were not achieved
int MemoryManager::swap_out(MMContext& context)
{
  message("\nSwapping out...");
  if( !context.enabled_address_translation )
    return 0;				// No virtual memory was occupied
  register int i;
  int swapped_something = 0;
  for(i=0; i<MMContext::process_virtual_space; i++)
  {
    VPN vpn = context.vm_map[i];
    if( vpn == NIL_vpn )
      continue;
    VirtualPage& vp = virtual_space[vpn];
    MMUContext::PageTableEntry& pte = context.page_table[i];
    if( !pte.valid )			// The page wasn't accessed by us yet
      continue;				// (if it was allocated at all)
    assert( pte.phys_page_no == vp.page_frame );
    virtual_space.unreference(vpn,pte.was_written);
    context.no_mapped_pages--;
    pte.valid = 0;
    pte.was_ref = pte.was_written = 0;
    swapped_something = 1;
  }
  assert(context.no_mapped_pages == 0);
  return swapped_something;
}

				// Run the LRU page replacement strategy 
				// to find out the least referenced page
				// and push it out
void MemoryManager::replace_LRU_page(MMContext& context)
{
  int max_counter = -1;
  int page_victim = -1;
  register int i;
  for(i=0; i<MMContext::process_virtual_space; i++)
    if(context.page_table[i].valid &&
       context.time_since_last_ref[i] > max_counter)
      max_counter = context.time_since_last_ref[i],
      page_victim = i;

  assert( page_victim >= 0 );
  VPN vpn = context.vm_map[page_victim];
  message("\nFound LRU page %d, time since last used %d",vpn,max_counter);

  MMUContext::PageTableEntry& pte = context.page_table[page_victim];
  assert( pte.valid );
  virtual_space.unreference(vpn,pte.was_written);
  context.no_mapped_pages--;
  pte.valid = 0;
  pte.was_ref = pte.was_written = 0;
}

				// Lock the page that contains the
				// specified virtual address. The page
				// should be already mapped
				// Return FALSE if failed
int MemoryManager::lock_loaded_page(MMContext& context,const Address addr)
{
  if( !context.enabled_address_translation )
    return 0;				// Virtual memory is disabled
  
  MMUContext::VirtualMemoryAddr va = *(MMUContext::VirtualMemoryAddr*)&addr;
  unsigned int page_no = va.page_no;
  assert( page_no < (unsigned)context.page_table_len );
  VPN vpn = context.vm_map[page_no];

  virtual_space.lock(vpn);
  return 1;
}

				// Unlock the page that contains the
				// specified virtual address. The page
				// should be already mapped & locked
void MemoryManager::unlock_page(MMContext& context,const Address addr)
{
  assert( context.enabled_address_translation );
  
  MMUContext::VirtualMemoryAddr va = *(MMUContext::VirtualMemoryAddr*)&addr;
  unsigned int page_no = va.page_no;
  assert( page_no < (unsigned)context.page_table_len );
  VPN vpn = context.vm_map[page_no];

  virtual_space.unlock(vpn);
}

				// Translate the virtual address to physical
				// one. The page has to be loaded.
				// Returns 0 if fails
Address MemoryManager::translate_va(MMContext& context,const Address addr)
{
  if( !context.enabled_address_translation )
    return addr;				// Virtual memory is disabled
  
  MMUContext::VirtualMemoryAddr va = *(MMUContext::VirtualMemoryAddr*)&addr;
  unsigned int page_no = va.page_no;
  if( page_no >= (unsigned)context.page_table_len )
    return 0;					// Out of range
  const MMUContext::PageTableEntry& pte = context.page_table[page_no];
  if( !pte.any_access )
    return 0;					// No access
  if( !pte.valid )
    return 0;					// Not mapped
  va.page_no = pte.phys_page_no;

  return *(Address *)&va;
}
