// This may look like C code, but it is really -*- C++ -*-
/*
 ************************************************************************
 *
 *			   UNT Virtual Machine
 *
 *		   Managing pages of real and virtual memory
 *
 ************************************************************************
 */

#include "mem_manager.h"
#include "myenv.h"

/*
 *========================================================================
 *		     Managing pages of physical memory
 */

PageFrame::PageFrame(void)
{
  id = key = 0;
}

/*
 *------------------------------------------------------------------------
 *		    Constructing the physical page table
 */

PageFrameTable::PageFrameTable(const int no_phys_pages)
	: no_page_frames(no_phys_pages)
{
  assure(no_page_frames > 4,"Number of physical pages must be at least 5");
  page_frames = new PageFrame[no_page_frames];

  register PFN i;
  for(i=0; i<no_page_frames; i++)
  {
    PageFrame& pf = (*this)[i];
    pf.id = i;
    pf.key = 0;
    if( i > sys_page_frame )
      freePFs.append(pf);
  }
}

/*
 *------------------------------------------------------------------------
 *		     Elementary Page Frame operations
 */

				// Get a Page Frame given its no
PageFrame& PageFrameTable::operator [] (const PFN pfn) const
{
  if( pfn == NIL_pfn )
    _error("FATAL: Requested page frame number is NIL");
  if( pfn >= no_page_frames )
    _error("FATAL: Page Frame Number %d is too big",pfn);
  return page_frames[pfn];
}

				// Print info about physical memory allocation
void PageFrameTable::dump_status(void) const
{
  message("\nPhysical memory: %d pages total, %d pages free\n",
	  no_page_frames,freePFs.q_no_elems());
}

				// Allocate a new physical page, return NIL
				// if no physical page is available
PFN PageFrameTable::allocate(void)
{
  if( freePFs.is_empty() )
    return NIL_pfn;
  return freePFs.get_from_head()->id;
}


				// Dispose of the given page
void PageFrameTable::dispose(const PFN pfn)
{
  assure( pfn > sys_page_frame, "May not dispose of a system page frame" );
  PageFrame& pf = (*this)[pfn];
  assert( pfn == pf.id );
  freePFs.append(pf);
}

/*
 *========================================================================
 *			Managing pages of virtual memory
 */


VirtualPage::VirtualPage(void)
{
  own_count = 0;
  ref_count = 0;
  page_frame = NIL_pfn;
  private_page = data_page = locked_page = 0;
}

/*
 *------------------------------------------------------------------------
 *		Constructing the virtual page table
 */

VirtualPageTable::VirtualPageTable
 (Memory& _physical_memory,const int no_phys_pages, const int no_virtual_pages)
	: no_pages(no_virtual_pages),
	  page_frames(no_phys_pages),
	  physical_memory(_physical_memory)
{
  assure(no_pages > 4,"Number of virtual pages must be at least 5");
  pages = new VirtualPage[no_pages];

  register VPN i;
  for(i=0; i<no_pages; i++)
  {
    VirtualPage& vp = (*this)[i];
    vp.id = i;
    vp.key = 0;
    freeVPs.append(vp);
  }
  no_locked_pages = 0;
}

/*
 *------------------------------------------------------------------------
 *		     Elementary Virtual Page operations
 */

				// Get a Virtual Page given its no
VirtualPage& VirtualPageTable::operator [] (const VPN vpn) const
{
  if( vpn == NIL_vpn )
    _error("FATAL: Requested virtual page number is NIL");
  if( vpn >= no_pages )
    _error("FATAL: Virtual Page Number %d is too big",vpn);
  return pages[vpn];
}

				// Print info about physical memory allocation
void VirtualPageTable::dump_status(void) const
{
  message("\nVirtual memory: %d pages total, %d pages free, %d pages locked\n",
	  no_pages,freeVPs.q_no_elems(),no_locked_pages);
  page_frames.dump_status();
}

				// Allocate a new virtual page, return NIL
				// if no page is available
VPN VirtualPageTable::allocate(void)
{
  if( freeVPs.is_empty() )
    return NIL_vpn;
  VirtualPage& vp =(*this)[freeVPs.get_from_head()->id];
  assert( vp.own_count == 0 && vp.ref_count == 0 );
  vp.own_count++;
  vp.page_frame = NIL_pfn;
  return vp.id;
}

				// Allocate a specified virtual page.
				// If it's already allocated, increment
				// the ownership count
VirtualPage& VirtualPageTable::allocate(VPN desired_vpn)
{
  VirtualPage& vp = (*this)[desired_vpn];
  if( vp.own_count++ == 0 )		// If the page has been free,
    freeVPs.dequeue(vp);		// dequeue it from the free list
  return vp;
}

				// Declare a page being private
				// Note, a page which is already private
				// or shared may not be declared private
void VirtualPageTable::declare_page_private(VPN vpn)
{
  VirtualPage& vp = allocate(vpn);
  assure( vp.own_count == 1 && vp.private_page == 0, 
	 "Illegal attempt to declare a page private" );
  vp.private_page = 1;
}

				// Copy virtual pages. system page frame is
				// used as a copy buffer
void VirtualPageTable::copy(const VPN to, const VPN from)
{
  const VirtualPage& from_vp = (*this)[from];
  
  if( from_vp.page_frame != NIL_pfn )	// 'From' page is already in the memory
    physical_memory.drum_write(from_vp.page_frame,to);
  else
  {					// Else read it from the drum
    physical_memory.drum_read(from,PageFrameTable::sys_page_frame);
    physical_memory.drum_write(PageFrameTable::sys_page_frame,to);
  }
}

				// Dispose of the given page
void VirtualPageTable::dispose(const VPN vpn)
{
  VirtualPage& vp = (*this)[vpn];
  assert( vpn == vp.id && vp.own_count > 0 );	// Consistency check

  if( vp.is_locked() )
    unlock(vpn);
  
  if( --vp.own_count > 0 )
    return;				// The page is still in use
  assert( vp.ref_count == 0 );		// The page should be unreferenced
  assert( vp.locked_page == 0 );	// The page should be unlocked
  if( vp.page_frame != NIL_pfn )
    page_frames.dispose(vp.page_frame),
    vp.page_frame = NIL_pfn,
    vp.ref_count = 0;

  if( vp.private_page )			// Page of exec image, should not
    return;				// be spoiled or disposed
  vp.data_page = 0;
  freeVPs.append(vp);
}

				// The page is being demanded, reference it
				// If necessary, the physical page frame is
				// allocated and the page is loaded.
				// NIL_pfn means the physical memory is
				// exhausted
PFN VirtualPageTable::reference(const VPN vpn)
{
  VirtualPage& vp = (*this)[vpn];
  assert( vpn == vp.id && vp.own_count > 0 );	// Consistency check
  message("\nMap: page %d, own count %d, prev. ref count %d",
	  vpn,vp.own_count,vp.ref_count);
  assert( vp.ref_count < vp.own_count );
  if( vp.ref_count > 0 )	// The page has already been referenced
  {
    assert( vp.page_frame != NIL_pfn );
    vp.ref_count++;
    message(": already mapped to page frame %d",vp.page_frame);
    return vp.page_frame;
  }
  if( (vp.page_frame = page_frames.allocate()) == NIL_pfn )
    return NIL_pfn;			// Physical space is exhausted

  message(": to be read into page frame %d",vp.page_frame);
  physical_memory.drum_read(vp.id,vp.page_frame);	// Load the page
  vp.ref_count++;
  return vp.page_frame;
}

				// The page is to be swapped out, unreference
void VirtualPageTable::unreference(const VPN vpn,const int to_write)
{
  VirtualPage& vp = (*this)[vpn];
  assert( vpn == vp.id && vp.own_count > 0 );	// Consistency check
  message("\nUnmap: virtual page %d, own count %d, ref count %d, "
	  "from page frame %d",
	  vpn,vp.own_count,vp.ref_count,vp.page_frame);
  assert( vp.ref_count > 0 && vp.page_frame != NIL_pfn );
  if( --vp.ref_count > 0 )
    return;				// The page is still referenced
  if( to_write )
    physical_memory.drum_write(vp.page_frame,vp.id);
  page_frames.dispose(vp.page_frame);
  vp.page_frame = NIL_pfn;
}

					// Lock the mapped page
					// Note, the page may not be shared
					// If the page is already locked,
					// locking it again has no effect
void VirtualPageTable::lock(const VPN vpn)
{
  VirtualPage& vp = (*this)[vpn];
  assert( vpn == vp.id && vp.own_count > 0 );	// Consistency check
  message("\nLock: virtual page %d, own count %d, ref count %d, "
	  "on page frame %d",
	  vpn,vp.own_count,vp.ref_count,vp.page_frame);
  assert( vp.ref_count > 0 && vp.page_frame != NIL_pfn );
  if( vp.is_locked() )
    return;
  assert( vp.ref_count == 1 );
  vp.locked_page = 1;
  vp.ref_count++;
  no_locked_pages++;
}

				// Unlock the mapped page
				// Unlocking the no page that hasn't been
				// locked has no effect
void VirtualPageTable::unlock(const VPN vpn)
{
  VirtualPage& vp = (*this)[vpn];
  assert( vpn == vp.id && vp.own_count > 0 );	// Consistency check
  message("\nUnlock: virtual page %d, own count %d, ref count %d, "
	  "on page frame %d",
	  vpn,vp.own_count,vp.ref_count,vp.page_frame);
  if( !vp.is_locked() )
    return;
  assert( --no_locked_pages >= 0 );
  assert( vp.ref_count > 0 && vp.page_frame != NIL_pfn );
  assert( vp.locked_page == 1 );
  vp.locked_page = 0;
  unreference(vpn,1);
}

				// Reserve the physical page for the system
				// Returns the physical address it starts
Address VirtualPageTable::reserve_physical_page()
{
  PFN pfn = page_frames.allocate(); 
  assure( pfn != NIL_pfn, "Failed to reserve a physical page for the system" );
  return pfn * Memory::pagesize;
}

